"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.modules import deserialize as deserialize
from synalinks.src.modules import serialize as serialize
from synalinks.src.modules.agents.function_calling_agent import (
    FunctionCallingAgent as FunctionCallingAgent,
)
from synalinks.src.modules.core.action import Action as Action
from synalinks.src.modules.core.branch import Branch as Branch
from synalinks.src.modules.core.decision import Decision as Decision
from synalinks.src.modules.core.generator import Generator as Generator
from synalinks.src.modules.core.identity import Identity as Identity
from synalinks.src.modules.core.input_module import Input as Input
from synalinks.src.modules.core.input_module import InputModule as InputModule
from synalinks.src.modules.core.not_module import Not as Not
from synalinks.src.modules.knowledge.embedding import Embedding as Embedding
from synalinks.src.modules.knowledge.update_knowledge import (
    UpdateKnowledge as UpdateKnowledge,
)
from synalinks.src.modules.merging.concat import Concat as Concat
from synalinks.src.modules.merging.concat import Concat as Concatenate
from synalinks.src.modules.merging.logical_and import And as And
from synalinks.src.modules.merging.logical_or import Or as Or
from synalinks.src.modules.merging.logical_xor import Xor as Xor
from synalinks.src.modules.module import Module as Module
from synalinks.src.modules.retrievers.entity_retriever import (
    EntityRetriever as EntityRetriever,
)
from synalinks.src.modules.retrievers.triplet_retriever import (
    TripletRetriever as TripletRetriever,
)
from synalinks.src.modules.synthesis.python_synthesis import (
    PythonSynthesis as PythonSynthesis,
)
from synalinks.src.modules.ttc.chain_of_thought import ChainOfThought as ChainOfThought
from synalinks.src.modules.ttc.self_critique import SelfCritique as SelfCritique
