"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.datasets.arcagi import default_instructions as default_instructions
from synalinks.src.datasets.arcagi import (
    get_arcagi1_evaluation_task_names as get_arcagi1_evaluation_task_names,
)
from synalinks.src.datasets.arcagi import (
    get_arcagi1_training_task_names as get_arcagi1_training_task_names,
)
from synalinks.src.datasets.arcagi import (
    get_arcagi2_evaluation_task_names as get_arcagi2_evaluation_task_names,
)
from synalinks.src.datasets.arcagi import (
    get_arcagi2_training_task_names as get_arcagi2_training_task_names,
)
from synalinks.src.datasets.arcagi import get_input_data_model as get_input_data_model
from synalinks.src.datasets.arcagi import get_output_data_model as get_output_data_model
from synalinks.src.datasets.arcagi import load_data as load_data
from synalinks.src.datasets.arcagi import plot_task as plot_task
